@@ -53,6 +53,7 @@ class RESTAssertion:
         return response.status_code == HTTPStatus.UNAUTHORIZED
 
 
+# noinspection PyPep8Naming
 class RESTClient(RESTAssertion):
     """ Main class for RestAPI """
 
@@ -66,11 +67,13 @@ class RESTClient(RESTAssertion):
                  auth_basic: bool = False,
                  auth_uri: str = '',
                  auth_payload: dict = None,
-                 auth_timeout: int = 7,
-                 logger_enabled: bool = True):
+                 connect_timeout: int = 5,
+                 read_timeout: int = 60,
+                 log_enabled: bool = True,
+                 log_level: str = 'INFO'):
 
         # Enable/disable logger
-        self.logger = logger(self.__class__.__name__, enabled=logger_enabled)
+        self.logger = logger('RESTClient', enabled=log_enabled, level=log_level)
 
         # Credentials and settings for api url
         self.username = username
@@ -81,49 +84,60 @@ class RESTClient(RESTAssertion):
         self.auth_uri = auth_uri
         self.auth_payload = auth_payload
         self.auth_basic = auth_basic
-        self.auth_timeout = auth_timeout
+        self.connect_timeout = connect_timeout
+        self.read_timeout = read_timeout
         self.header = header or get_header_default()  # Default header
 
     def __str__(self):
+        msg_print = (f'Username: {self.username}\n'
+                     f'Password: {self.password}\n'
+                     f'Base URL: {self.base_url}\n')
 
-        return f'Username: {self.username}\n' \
-               f'Password: {self.password}\n' \
-               f'Base URL: {self.base_url}\n'
+        return msg_print
 
     @property
     def base_url(self):
-        """Create tested URL"""
+        """Create tested URL. Add port if it provided"""
 
-        return f'{self.protocol}://{self.host}:{self.port}' if self.port else f'{self.protocol}://{self.host}'
+        base_url = f'{self.protocol}://{self.host}:{self.port}' if self.port else f'{self.protocol}://{self.host}'
 
-    def requested_url(self, url):
-        """ Return full URL if link specified contains 'http' """
+        return base_url
 
-        if 'http' in url:
-            return url
-        elif self._validate_url(self.base_url):
-            return urljoin(self.base_url, url)
-        raise SyntaxError(f'URL does not specified. Used: "{url}" only. Specify full URL or use constructor')
+    def requested_url(self, url) -> str:
+        """Return full URL if link specified starts with <http>"""
+
+        try:
+            requested_url = url if url.startswith('http') else urljoin(self.base_url, url)
+        except Exception as err:
+            self.logger.exception(f'URL does not specified. Used: "{url}" only. Specify full URL or use constructor')
+            raise err
+
+        return requested_url
 
     @staticmethod
     def _validate_url(url):
-        return re.match(r'http[s]?://[\w/.]+', url)
+        return re.match(r'https?://[\w/.]+', url)
 
-    def encode_to_base64(self, string: str):
-        """Encode string to base64
+    def encode_to_base64(self, string: str) -> str:
+        """Encode string to base64. Do not use it during creating class instance (logger won't work)
 
-        :param string:
-        :return:
+        Args:
+            string: Some text to encode.
+
+        Returns:
+            Base64 encoded string.
         """
 
         try:
             str_text = string.encode()
             base64_encoded = base64.b64encode(str_text)
-            return base64_encoded.decode()
         except AttributeError as err:
-            self.logger.exception(f'Cannot encode specific data ({string}, {type(string)})')
+            self.logger.error(f'Cannot encode specific data ({string}, {type(string)})')
             raise err
 
+        result = base64_encoded.decode()
+        return result
+
     def get_token(self) -> str:
         """Get auth (Bearer) token"""
 
@@ -142,6 +156,7 @@ class RESTClient(RESTAssertion):
             self.header = get_header_default()
 
         try:
+            # noinspection PyAttributeOutsideInit
             self.token = self.get_token()
             self.header['Authorization'] = self.token
             return self.header
@@ -163,8 +178,7 @@ class RESTClient(RESTAssertion):
                      verify: bool = False,
                      custom_header: dict = None,
                      cookies=None,
-                     request_timeout: int = 60):
-
+                     read_timeout: int = None):
         """Send common REST request
 
         To upload file use:
@@ -172,19 +186,25 @@ class RESTClient(RESTAssertion):
         with open(files, 'rb') as f:
              files = {'licenseFile': (License.name, f)}
 
-        :param cookies:
-        :param method: GET, POST, DELETE
-        :param url: unified identifier, self.url (BASE) + url
-        :param query_params: = Query parameters in the url
-        :param data: json data
-        :param files: "files: {'licenseFile': (License.name, f)}"
-        :param verify: Bool
-        :param custom_header: Use specified header
-        :param request_timeout: Time in sec.
-        :return: requests.request('POST', url=url, headers=HEADER, files=files, verify=False)
+        Args:
+            method: GET, POST, DELETE
+            url: unified identifier, self.url (BASE) + url
+            data: json data
+            files: "files: {'licenseFile': (License.name, f)}"
+            query_params: Query parameters in the url
+            cookies:
+            verify: Ignore SSL verification?
+            custom_header: Use specified header
+            read_timeout: Read timeout in sec.
 
+        Returns:
+            requests.request('POST', url=url, headers=HEADER, files=files, verify=False)
         """
 
+        # Get timeouts
+        read_timeout_ = self.read_timeout if read_timeout is None else read_timeout
+        common_timeout = self.read_timeout, read_timeout_
+
         # Set auth method
         if self.auth_payload is not None:
             self.header = self.get_auth_header(header=custom_header)
@@ -202,9 +222,8 @@ class RESTClient(RESTAssertion):
             link += '?' + urlencode(query_params)
 
         method = method.upper()
-
-        assert method in ['GET', 'HEAD', 'DELETE', 'POST', 'PUT', 'PATCH', 'OPTIONS'], \
-            'Specified methods is not compatible.'
+        methods_available = 'GET', 'HEAD', 'DELETE', 'POST', 'PUT', 'PATCH', 'OPTIONS'
+        assert method in methods_available, 'Specified methods is not compatible.'
 
         if data and files:
             self.logger.error('Data parameter cannot be used with files parameter simultaneously.')
@@ -223,7 +242,7 @@ class RESTClient(RESTAssertion):
             if method in ['POST', 'PUT', 'PATCH', 'OPTIONS', 'DELETE']:
                 if data and not files:
 
-                    self.logger.debug(f'{self.header = }, {cookies = }, {request_timeout = }')
+                    self.logger.debug(f'{self.header = }, {cookies = }, {common_timeout = }')
                     self.logger.info('PAYLOAD:\n' + json.dumps(data, indent=4))
 
                     response = requests.request(method=method,
@@ -233,13 +252,13 @@ class RESTClient(RESTAssertion):
                                                 json=data,
                                                 cookies=cookies,
                                                 verify=verify,
-                                                timeout=request_timeout)
+                                                timeout=common_timeout)
 
                 elif files:
                     files_header = self.header.copy()
                     del files_header['Content-Type']
 
-                    self.logger.debug(f'{files_header = }, {cookies = }, {request_timeout = }')
+                    self.logger.debug(f'{files_header = }, {cookies = }, {common_timeout = }')
                     self.logger.info('PAYLOAD:\n' + json.dumps(data, indent=4))
 
                     response = requests.request(method=method,
@@ -249,10 +268,10 @@ class RESTClient(RESTAssertion):
                                                 cookies=cookies,
                                                 files=files,
                                                 verify=verify,
-                                                timeout=request_timeout)
+                                                timeout=common_timeout)
 
                 elif not files and not data:
-                    self.logger.debug(f'{self.header = }, {cookies = }, {request_timeout = }')
+                    self.logger.debug(f'{self.header = }, {cookies = }, {common_timeout = }')
 
                     response = requests.request(method=method,
                                                 url=link,
@@ -260,12 +279,12 @@ class RESTClient(RESTAssertion):
                                                 headers=self.header,
                                                 cookies=cookies,
                                                 verify=verify,
-                                                timeout=request_timeout)
+                                                timeout=common_timeout)
 
             # For 'GET', 'HEAD' request
             else:
                 # For usual 'GET' without query params
-                self.logger.debug(f'{self.header = }, {cookies = }, {request_timeout = }')
+                self.logger.debug(f'{self.header = }, {cookies = }, {common_timeout = }')
 
                 response = requests.request(method=method,
                                             url=link,
@@ -273,13 +292,15 @@ class RESTClient(RESTAssertion):
                                             cookies=cookies,
                                             headers=self.header,
                                             verify=verify,
-                                            timeout=request_timeout)
-
+                                            timeout=common_timeout)
+        except requests.exceptions.ReadTimeout as err:
+            self.logger.exception(f'The request took too long. Read timed out. (read timeout={common_timeout[1]})')
+            raise err
         except BaseException as err:
-            self.logger.exception('Something wend wrong.')
+            self.logger.exception('Something went wrong.')
             raise err
 
-        self.logger.info(f'[Response] {response.status_code}: "{response.text}"')
+        self.logger.info(f'<- {response.status_code}: "{response.text}"')
 
         return response
 
@@ -322,14 +343,14 @@ class RESTClient(RESTAssertion):
             url: str = '',
             query_params: dict = None,
             cookies=None,
-            request_timeout: int = None,
+            read_timeout: int = None,
             header: dict = None):
 
         return self.send_request(method='GET',
                                  url=url,
                                  query_params=query_params,
                                  cookies=cookies,
-                                 request_timeout=request_timeout,
+                                 read_timeout=read_timeout,
                                  custom_header=header)
 
     # noinspection PyPep8Naming
@@ -339,7 +360,7 @@ class RESTClient(RESTAssertion):
              file=None,
              query_params: dict = None,
              cookies=None,
-             request_timeout: int = None,
+             read_timeout: int = None,
              header: dict = None):
 
         return self.send_request(method='POST',
@@ -348,7 +369,7 @@ class RESTClient(RESTAssertion):
                                  data=data,
                                  files=file,
                                  cookies=cookies,
-                                 request_timeout=request_timeout,
+                                 read_timeout=read_timeout,
                                  custom_header=header)
 
     # noinspection PyPep8Naming
@@ -358,7 +379,7 @@ class RESTClient(RESTAssertion):
             files='',
             query_params: dict = None,
             cookies=None,
-            request_timeout: int = None,
+            read_timeout: int = None,
             header: dict = None):
 
         header_original = self.header.copy()
@@ -372,7 +393,7 @@ class RESTClient(RESTAssertion):
                                  files=files,
                                  query_params=query_params,
                                  cookies=cookies,
-                                 request_timeout=request_timeout,
+                                 read_timeout=read_timeout,
                                  custom_header=header)
 
     def DELETE(self,
@@ -380,7 +401,7 @@ class RESTClient(RESTAssertion):
                data: dict = '',
                query_params: dict = None,
                cookies=None,
-               request_timeout: int = None,
+               read_timeout: int = None,
                header: dict = None):
 
         return self.send_request(method='DELETE',
@@ -388,7 +409,7 @@ class RESTClient(RESTAssertion):
                                  query_params=query_params,
                                  data=data,
                                  cookies=cookies,
-                                 request_timeout=request_timeout,
+                                 read_timeout=read_timeout,
                                  custom_header=header)
 
     def is_host_available(self, port: int = 0, timeout: int = 5) -> bool:
@@ -453,7 +474,7 @@ class AuthToken:
 
         try:
             payload = self.client.auth_payload
-            timeout = self.client.auth_timeout
+            timeout = (self.client.connect_timeout, self.client.read_timeout)
             response = requests.post(self.auth_url, json=payload, verify=False, timeout=timeout)
             response_json = response.json()
             return response_json['token']
